#include <math.h>
#include <time.h>
#include "Puzzle.h"

using namespace std;

Puzzle::Puzzle()
{
	n = n_orig = 0;
	zeroPosition = 0;
}

Puzzle::Puzzle(const Puzzle& p2)
{
	n = n_orig = p2.GetN();
	tiles = tiles_orig = p2.tiles;
	CountZeroPosition();

}

Puzzle::~Puzzle()
{ }

void Puzzle::ResetTiles()
{
	this->tiles.clear();
	return;
}

void Puzzle::ResetTiles(unsigned int n)
{
	this->tiles.clear();
	this->tiles.resize(n*n,0);
	return;
}

unsigned int Puzzle::ManhattanDist() const
{
	unsigned int dist = 0;
	int radek_1, radek_2, sloupec_1, sloupec_2 = 0;
	for(unsigned i=0; i<tiles.size();++i)
	{
		if(tiles[i]==0)
				continue;

		PositionToCoord(i,radek_1,sloupec_1);
		PositionToCoord(GetGoalPosition(tiles[i]),radek_2,sloupec_2);
		dist += abs(radek_1-radek_2) + abs(sloupec_1-sloupec_2);
	}

	return dist;
}

bool Puzzle::CanTrim() const
{
	if(tiles.empty())
		return false;

	bool ok = true;
	for(unsigned i=0; i<n; ++i)
	{
		ok = ok && (tiles[i]==i+1) && ( tiles[i*n]==(i*n)+1 );
	}
	return ok;

}

bool Puzzle::TrimBoard()
{
	if(!CanTrim())
	{
		return false;
	}
	else
	{
		// muzu to udelat
		if((this->n>1) && (tiles.size()==n*n))
		{
			// udela se nova "deska":
			vector<int> tiles_new;
			int tile, tile_new = 0;
			for(unsigned index=n+1;index<n*n;++index)	// staci jit od n+1, pze ty predtim jsou oriznuty
			{
				tile = tiles[index];
				if(index%n==0)
				{
					// nedelat nic, pze to je 1. sloupec
				}
				else
				{
					tile_new =  (tile == 0) ? 0 : ( tile - n - ((tile-1)/n) );		
					// -n za 1.radu, (tile-1)/n za kazdy sloupec (ta -1 je tam proto, ze kameny jdou od 1, ne od 0)
					tiles_new.push_back(tile_new);
				}
			}

			tiles = tiles_new;
			n-=1;
		}
		this->CountZeroPosition();
		return true;
	}
}



//
bool Puzzle::DoMove(const char move)
{
	switch(move)
	{
		case 'U':
					if(zeroPosition>=n)
					{
						tiles[zeroPosition]=tiles[zeroPosition-n];
						tiles[zeroPosition-n]=0;
						zeroPosition=zeroPosition-n;
						return true;
					}
					break;					
		case 'D':
					if(zeroPosition<(n*n-n))
					{
						tiles[zeroPosition]=tiles[zeroPosition+n];
						tiles[zeroPosition+n]=0;
						zeroPosition=zeroPosition+n;
						return true;
					}
					break;
		case 'L':
					if(zeroPosition%n!=0)
					{
						tiles[zeroPosition]=tiles[zeroPosition-1];
						tiles[zeroPosition-1]=0;
						zeroPosition--;
						return true;
					}
					break;
		case 'R':
					if(zeroPosition%n!=(n-1))
					{
						tiles[zeroPosition]=tiles[zeroPosition+1];
						tiles[zeroPosition+1]=0;
						zeroPosition++;
						return true;
					}
					break;
		default:	return false;

	}
	return false;
}

bool Puzzle::ApplySequence(const std::string sequence)
{
	bool ok = true;
	for(unsigned i=0;i<sequence.length();++i)
	{
		if(DoMove(sequence[i]))
		{
			// ok = ok;
		}
		else
		{
			ok = false;
			cout << "\nWarning: can't use this move - " << sequence[i] << " (ApplySequence("<<sequence<<"))";
		}
	}
	return ok;
}



unsigned int Puzzle::CountZeroPosition()
{
	zeroPosition = 0;
	for(unsigned i=0;i<tiles.size();++i)
	{
		if(tiles[i]==0)
		{
			zeroPosition = i;
			return i;
		}
	}
	return 0;
}

unsigned int Puzzle::CountZeroPosition() const
{
	for(unsigned i=0;i<tiles.size();++i)
	{
		if(tiles[i]==0)
		{
			return i;
		}
	}
	return 0;
}

bool Puzzle::IsSolvable() const
{
	if(!n || tiles.empty())
			return false;

	unsigned int trueZeroPosition  = 0; 
	if(tiles[zeroPosition]!=0)
	{
		trueZeroPosition = CountZeroPosition();
		cout << "Warning - in function IsSolvable: invalid zero position";
	}
	else
	{
		trueZeroPosition = zeroPosition;
	}

	int inverzi = 0;
	int kamen = -1;

	// nejprve potreba situaci na desce "znormovat" -> posunout nulu doprava dolu
	vector<int> tilesNorm = tiles;
	int zeroRow,zeroCol = 0;
	unsigned int zeroPositionNorm = CountZeroPosition();
	PositionToCoord(zeroPositionNorm,zeroRow,zeroCol);
	// posun doprava
	for(unsigned i=zeroCol; i<n-1; ++i)
	{
		tilesNorm[zeroPositionNorm] = tilesNorm[zeroPositionNorm+1];
		tilesNorm[zeroPositionNorm+1] = 0;
		zeroPositionNorm++;
	}
	// posun dolu
	for(unsigned i=zeroRow; i<n-1; ++i)
	{
		tilesNorm[zeroPositionNorm] = tilesNorm[zeroPositionNorm+n];
		tilesNorm[zeroPositionNorm+n] = 0;
		zeroPositionNorm += n;
	}

	// spocitat pocet inverzi
	for(unsigned i=0;i<tilesNorm.size()-1;++i)				// na posl. polizkuje mezera, tu pocitat nemusim
	{
		for(unsigned j=i+1;j<tilesNorm.size()-1;++j)
		{
			if(tilesNorm[i]>tilesNorm[j])
				++inverzi;
		}
	}
	
	return (inverzi%2==0);

}

void Puzzle::PositionToCoord(int pos, int &row, int &col) const
{
	if(n>0)
	{
		col = pos % n;
		row = (pos / n);
		return;
	}
	else
	{
		printf("Error in function PositionToCoord( %d ) : n is set to 0",pos);
		return;
	}
}

bool Puzzle::RandomMove()
{
	if(tiles.empty() || n==0)
		return false;

	char move = 'X';
	do
	{
		int rand4 = (int)( ((double) rand() / (double) RAND_MAX) * 4);
		switch(rand4)
		{
			case 0:		move = 'U';
						break;					
			case 1:		move = 'D';
						break;
			case 2:		move = 'L';
						break;
			case 3:		move = 'R';
						break;

		}
	}
	while(!this->DoMove(move));

	return true;

}

// I/O
bool Puzzle::LoadFromSbp(std::string filename)
{
	const char* fname = filename.c_str();

	TiXmlDocument doc( fname );
	bool loadOkay = doc.LoadFile();
	if (loadOkay)
	{
		if(this->LoadXml(&doc))
		{
			this->n_orig = this->n;
			this->tiles_orig = this->tiles;
			CountZeroPosition();
		}
	}
	else
	{
		printf("Failed to load file: %s\n",fname);
		return false;
	}

	return true;
}


bool Puzzle::SaveToSbp(std::string filename)
{
	const char* fname = filename.c_str();

	TiXmlDocument doc;  

	// 1) deklarace
	TiXmlDeclaration* decl = new TiXmlDeclaration( "1.0", "", "" );  
	doc.LinkEndChild( decl );



	// 2) hlavni element
	TiXmlElement* puzzleElem = new TiXmlElement("SBPuzzle");
	puzzleElem->SetAttribute("Width",this->n);
	puzzleElem->SetAttribute("Height",this->n);
	puzzleElem->SetAttribute("IndirectPushAllowed","False");
	puzzleElem->SetAttribute("MultiDirectionalMoves","False");
	puzzleElem->SetAttribute("CompactLines","False");


	// 3) kameny
	TiXmlElement pieceElem = TiXmlElement("Piece");
	pieceElem.SetAttribute("Width",1);
	pieceElem.SetAttribute("Height",1);
	pieceElem.SetAttribute("Color","-16777201");
	pieceElem.SetAttribute("VSliding","True");
	pieceElem.SetAttribute("HSliding","True");

	// 3a) startovni pozice
	TiXmlElement initialPosElem = TiXmlElement("InitialPos");
	int radek, sloupec=0;

	// 3b cilova pozice
	TiXmlElement goalPosElem = TiXmlElement("GoalPos");
	
	// 4) projit pres vsechny kameny
	for(unsigned i=0;i<this->tiles.size();++i)
	{
		if(tiles[i]==0)
				continue;

		pieceElem.SetAttribute("Caption",tiles[i]);
		pieceElem.Clear();

		// start:
		PositionToCoord(i,radek,sloupec);
		initialPosElem.SetAttribute("Line",radek);
		initialPosElem.SetAttribute("Col",sloupec);
		pieceElem.InsertEndChild(initialPosElem);

		// cil:
		PositionToCoord(this->GetGoalPosition(tiles[i]),radek,sloupec);
		goalPosElem.SetAttribute("Line",radek);
		goalPosElem.SetAttribute("Col",sloupec);
		pieceElem.InsertEndChild(goalPosElem);

		puzzleElem->InsertEndChild(pieceElem);
	}

	doc.LinkEndChild(puzzleElem);
	
	// 4) ulozit soubor
	return doc.SaveFile(fname);
}

bool Puzzle::DirectLoad(std::string filename)
{
	ifstream fin(filename.c_str());
	int i;
	if(fin)
	{
		tiles.clear();
		while(fin >> i)
		{
			// cout << "\n cislo : " << i << endl;
			tiles.push_back(i);
		}
		// ted jestli to ma spravnou velikost:
		double n_d = sqrt((double)tiles.size());
		n = (unsigned int) n_d;
		if(n*n == tiles.size())
		{
			// ok
			n_orig = n;
			tiles_orig = tiles;
			CountZeroPosition();
		}
		else
		{
			cout << "Error - board isn't n x n!";
		}
	}

	return true;
}



void Puzzle::Print() const
{
	if(n==0)
	{
		printf("No puzzle to be print!");
		return;
	}
	else
	{
		// system("cls");
		cout << "\n\nSize: " << n << "x" << n << endl;
		cout << "Manhattan distance: " << this->ManhattanDist() << endl;
		cout << "Solvable: ";
		if(IsSolvable())
			cout << "Yes" << endl;
		else
			cout << "NO!" << endl;
		
		if(this->n>20)
		{
			cout << "Can only print boards with size <=20";
			return;
		}	
		cout << "\nTiles: " << endl;

		for(unsigned i=0;i<this->tiles.size();++i)
		{
			if(i%n==0)
				cout << "\n";
			
			printf("%4d",tiles[i]);
		}
	}


}

bool Puzzle::LoadXml(TiXmlDocument* doc)
{
	// nastavit
	if(!doc)
		return false;

	TiXmlElement* pElem;
	TiXmlElement* pInitialPosElem;
	TiXmlHandle hRoot(0);
	int rozmer=0;


	// projit childs
	// 1) element SBPuzzle -> parametry Width a Height
	hRoot = doc->FirstChildElement("SBPuzzle");
	if(pElem=hRoot.ToElement())
	{
		if(!pElem->Attribute("Width",&rozmer))
			return false;
		else
		{
			this->n = rozmer;
			ResetTiles(rozmer);
		}
	}
	else
		return false;


	// 2) vsechny deti elementu SBPuzzle jmenem Piece
	int cislo = 0;	// cislo, ktere je na kameni napsano
	int rada, sloupec = 0;
	pElem = hRoot.FirstChildElement("Piece").ToElement();
	while(pElem)
	{
		if(!pElem->Attribute("Caption",&cislo))
		{
			printf("Failed to load from xml: Element 'Piece' has no attribute 'Caption'");
			return false;
		}
		else
		{
			// zjistit startovni pozici (tu cilovou znam, ta je vzdy stejna)
			if(!(pInitialPosElem=pElem->FirstChildElement("InitialPos")))
			{
				printf("Failed to load from xml: Element 'Piece' %d has no element 'InitialPos'",cislo);
				return false;
			}
			else
			{	
				// nacist radek a sloupec
				if(pInitialPosElem->Attribute("Line",&rada) && pInitialPosElem->Attribute("Col",&sloupec))
				{
					// cout << cislo << ":" << rada << " " << sloupec << endl;
					
					// ulozit na spravnou pozici ve vektoru:
					unsigned pozice = rada * this->n + sloupec;
					if(pozice<this->tiles.size())
					{
						this->tiles[pozice] = cislo;						
					}
					else
					{
						printf("Failed to load from xml: tile %d has wrong initial position",cislo);
						return false;
					}

				}
				else
				{
					printf("Failed to load from xml: Element 'Piece' %d has no element 'GoalPos' with 'Line' and 'Col'",cislo);
					return false;
				}
			}
		}

		pElem = pElem->NextSiblingElement("Piece");
	}


	return true;
}


// Gett

int Puzzle::operator [](unsigned int pos) const
{
	if(pos<tiles.size())
	{
		return tiles[pos];
	}
	else
		return -1;
}

unsigned int Puzzle::GetPosition(int tile) const
{
	unsigned int pos = 0;
	for(unsigned i=0; i<tiles.size(); ++i)
	{
		if(tiles[i]==tile)
		{
			pos = i;
			break;
		}
	}
	return pos;

}
unsigned int Puzzle::GetGoalPosition(int tile) const
{
	if(n>0 && tile>0)
	{
		tile-=1;	// pocita se od 0
		return (tile % n) + n * (tile / n);
	}
	else
	{
		printf("Error in function GetGoalPosition( %d ) : n=%d",tile,n);
		return 0;
	}

}

unsigned int Puzzle::GetN() const
{
	return this->n;
}
void Puzzle::GetNeighbours(unsigned int pos, std::vector<SOUSED>& sousedi) const
{
	sousedi.clear();

	// case 'U':
	if(pos>=n)
	{
		sousedi.push_back( make_pair(tiles[pos-n],pos-n) );
	}
	// case 'D':
	if(pos<(n*n-n))
	{
		sousedi.push_back( make_pair(tiles[pos+n],pos+n) );
	}
	// case 'L':
	if(pos%n!=0)
	{
		sousedi.push_back( make_pair(tiles[pos-1],pos-1) );
	}
	// case 'R':
	if(pos%n!=(n-1))
	{
		sousedi.push_back( make_pair(tiles[pos+1],pos+1) );
	}
	
}
// Sett
unsigned int Puzzle::SetN(unsigned int n2)
{
	return this->n = n2;
}
unsigned int Puzzle::SetZeroPosition(unsigned int zeroPosition2)
{
	return this->zeroPosition = zeroPosition2;
}

void Puzzle::SetTiles(const vector<int>& newTiles)
{
	tiles = tiles_orig = newTiles;
	CountZeroPosition();
}

const char DetermineDirection(int pos1, int pos2, unsigned int n)
{
	// z p1 na p2 jaky je to smer
	if(pos1==pos2-1)
		return 'R';
	else if(pos1==pos2+1)
		return 'L';
	else if(pos1==pos2-n)
		return 'D';
	else if(pos1==pos2+n)
		return 'U';
	else
		return 'X';

}